/*
 * Copyright (c) 2003-2004 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */
#include "oscompat.h"
#include "typedefs.h"
#include "osservices.h"

#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>
#include <linux/pci.h>
#include <asm/pgtable.h>

//#define OS_DEBUG_MEM

#define DLDR_MEM_CONTIGUOUS 0x00000001
#define DLDR_MEM_NONCACHED  0x00000002
#define DLDR_MEM_EXECUTABLE 0x00000004

static int pre_vmalloc_table_size;
static int pre_vmalloc_block_size;

static spinlock_t pre_vmalloc_lock __attribute__((unused)) = SPIN_LOCK_UNLOCKED;
static struct { void *ptr; BOOLEAN inuse; } *pre_vmalloc;

__shimcall__
void *OsAllocateFlags(unsigned size, unsigned allocflags)
{
	PVOID pMem;
   
	if(allocflags & DLDR_MEM_EXECUTABLE) {
#if defined(_PAGE_KERNEL_EXEC)
#define USE_PAGE_KERNEL_EXEC __pgprot(_PAGE_KERNEL_EXEC)
#elif defined(PAGE_KERNEL_EXEC)
#define USE_PAGE_KERNEL_EXEC PAGE_KERNEL_EXEC
#elif defined(PAGE_KERNEL_EXECUTABLE)
#define USE_PAGE_KERNEL_EXEC PAGE_KERNEL_EXECUTABLE
#endif
#ifdef USE_PAGE_KERNEL_EXEC
		pMem = NULL;

		if(OsGetIrql() == OSIRQL_PASSIVE)
			pMem = __vmalloc(size, GFP_KERNEL | __GFP_HIGHMEM, USE_PAGE_KERNEL_EXEC);

		if(pMem) {
#ifdef OS_DEBUG_MEM
			printk(KERN_DEBUG "%s(%u, DLDR_MEM_EXECUTABLE): %p\n", __FUNCTION__, size, pMem);
#endif
			memset(pMem, 0, size); /* caller expects memory to be initialized to zero */
		} else {
			printk(KERN_ERR "%s(%u, DLDR_MEM_EXECUTABLE): failed, OsIrql=%d\n", __FUNCTION__, size, OsGetIrql());
		}
		return pMem;
#else
		return OsAllocate(size);
#endif
	}

	pMem = kmalloc(size, (OsGetIrql() == OSIRQL_PASSIVE) ? GFP_KERNEL : GFP_ATOMIC);
	if(!pMem) {
		if(!(allocflags & DLDR_MEM_CONTIGUOUS) && (OsGetIrql() == OSIRQL_PASSIVE)) {
			/* try vmalloc since kmalloc is generally limited to 64k */
			pMem = vmalloc(size);
			//printk(KERN_DEBUG "%s(%u): vmalloc returned %p\n", __FUNCTION__, size, pMem);
		}
		if(!pMem && !(allocflags & DLDR_MEM_CONTIGUOUS) && (size <= pre_vmalloc_block_size)) {
			unsigned long flags;
			int i;

			spin_lock_irqsave(&pre_vmalloc_lock, flags);
			for(i = 0; i < pre_vmalloc_table_size; i++) {
				if(!pre_vmalloc[i].inuse) {
					pMem = pre_vmalloc[i].ptr;
					pre_vmalloc[i].inuse = 1;
					break;
				}
			}
			spin_unlock_irqrestore(&pre_vmalloc_lock, flags);

#if 0
			if(pMem) {
				printk(KERN_ERR "%s(%u, 0x%x): from pre_vmalloc, OsIrql=%d %p\n", __FUNCTION__, size, allocflags, OsGetIrql(), pMem);
			}
#endif
		}
	}

	if(pMem) {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s(%u, 0x%x): %p\n", __FUNCTION__, size, allocflags, pMem);
#if 0
		if((size == 21) || (size == 85))
			OsDumpStack();
#endif
#endif
		memset(pMem, 0, size); /* caller expects memory to be initialized to zero */
	} else {
		printk(KERN_ERR "%s(%u, 0x%x): failed, OsIrql=%d\n", __FUNCTION__, size, allocflags, OsGetIrql());
	}

	return pMem;
}
 
__shimcall__
void OsFree(PVOID pMem)
{
	unsigned long flags;
	int i;

	if(!pMem) {
		printk(KERN_ERR "%s: NULL\n", __FUNCTION__);
		return;
	}

	spin_lock_irqsave(&pre_vmalloc_lock, flags);
	for(i = 0; i < pre_vmalloc_table_size; i++) {
		if(pre_vmalloc[i].ptr == pMem) {
			//printk(KERN_DEBUG "%s: pre_vmalloc return %p\n", __FUNCTION__, pMem);
			pre_vmalloc[i].inuse = 0;
			spin_unlock_irqrestore(&pre_vmalloc_lock, flags);
			return;
		}
	}
	spin_unlock_irqrestore(&pre_vmalloc_lock, flags);

	if((pMem >= (PVOID)VMALLOC_START) && (pMem < (PVOID)VMALLOC_END)) {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s: vfree %p\n", __FUNCTION__, pMem);
#endif
		if(OsGetIrql() != OSIRQL_PASSIVE) {
			printk(KERN_WARNING "%s: calling vfree(%p) in wrong context, OsIrql=%d\n", __FUNCTION__, pMem, OsGetIrql());
		}
		vfree(pMem);
	} else {
#ifdef OS_DEBUG_MEM
		printk(KERN_DEBUG "%s: %p\n", __FUNCTION__, pMem);
#endif
		kfree(pMem);
	}
}

__shimcall__
void *OsDmaAllocate(HANDLE DevNode, unsigned size, unsigned long *physAddr)
{
	dma_addr_t dma_addr;
	void *virt_addr;

	virt_addr = pci_alloc_consistent((struct pci_dev *)DevNode, size, &dma_addr);

	*physAddr = dma_addr;

	return virt_addr;
}

__shimcall__
void OsDmaFree(HANDLE DevNode, unsigned size, void *vAddr, unsigned long physAddr)
{
	pci_free_consistent((struct pci_dev *)DevNode, size, vAddr, physAddr);
}

__shimcall__
unsigned long OsAllocateContiguousMemory(unsigned size, BOOL isDMA, unsigned long *physAddr)
{
	unsigned long virt;

	virt = __get_free_pages(GFP_KERNEL | (isDMA ? GFP_DMA : 0), get_order(size));
	if (virt) {
		memset((void *)virt, 0, size);
		*physAddr = virt_to_phys((void *)virt);
	}

	return virt;
}

__shimcall__
void OsFreeContiguousMemory(unsigned long virtAddr, unsigned size)
{
	free_pages(virtAddr, get_order(size));
}

__shimcall__
int OsMemInit(void)
{
	return 0;
}

__shimcall__
int OsMemPreAllocate(int tableSize, int blockSize)
{
	int i;

	if(pre_vmalloc)
		return -EBUSY;

	if(tableSize) {
		pre_vmalloc = kmalloc(tableSize * sizeof(*pre_vmalloc), GFP_KERNEL);
		if(!pre_vmalloc)
			return -ENOMEM;
		memset(pre_vmalloc, '\0', tableSize * sizeof(*pre_vmalloc));

		pre_vmalloc_table_size = tableSize;
		pre_vmalloc_block_size = blockSize;

		for(i = 0; i < pre_vmalloc_table_size; i++) {
			pre_vmalloc[i].ptr = vmalloc(pre_vmalloc_block_size);
			if(!pre_vmalloc[i].ptr) {
				printk(KERN_ERR "%s(): pre_vmalloc slot %d allocation failed\n", __FUNCTION__, i);
				break;
			}
		}
	}

	return 0;
}

__shimcall__
void OsMemExit(void)
{
	int i;

	for(i = 0; i < pre_vmalloc_table_size; i++) {
		if(pre_vmalloc[i].ptr) {
			if(pre_vmalloc[i].inuse) {
				printk(KERN_ERR "%s(): pre_vmalloc_free slot %d (%p) still active\n", __FUNCTION__, i, pre_vmalloc[i].ptr);
			}
			vfree(pre_vmalloc[i].ptr);
			pre_vmalloc[i].ptr = NULL;
		}
	}
}

